from .core import current_app, plugin, shutdown, stop_all, App, AppStarted
from .provider import get, provide
from .configuration import config
from .event import Event, on, emit
from .loop import loop
from .cli import command
from .utils import require, uuid, tri_wave, float_range



import logging
import signal
from typing import Callable, Optional

logger = logging.getLogger(__name__)
exiting = False
sigint_callback: Optional[Callable[[int], None]] = None


def _sigint_handler(signum, frame):
    logger.info(f'Received sigint: {signum}')

    global exiting
    if exiting:
        logger.warning('Double sigint received, forcing exit')
        exit(1)

    exiting = True
    if sigint_callback is not None:
        sigint_callback(signum)


signal.signal(signal.SIGINT, _sigint_handler)



from dataclasses import dataclass
from threading import Thread
from typing import Type, Callable

import easy_pysi as ez
from easy_pysi.plugin import Plugin


class Event:
    pass


@dataclass
class EventSubscriber:
    callback: Callable[[Event], str]
    event_type: Type[Event]
    asynchronous: bool


class EventPlugin(Plugin):
    subscribers: list[EventSubscriber] = []

    def start(self):
        self.subscribers = [
            subscriber
            for subscriber in _ALL_SUBSCRIBERS
            if self.app.is_available(subscriber.callback)
        ]


_ALL_SUBSCRIBERS: list[EventSubscriber] = []


def on(*event_types: Type[Event], asynchronous=False):
    # TODO: auto detect async function?
    def decorator(func):
        for event_type in event_types:
            subscriber = EventSubscriber(func, event_type, asynchronous)
            _ALL_SUBSCRIBERS.append(subscriber)
        return func
    return decorator


def emit(event: Event):
    event_type = type(event)

    # Synchronous event
    for subscriber in _get_subscribers(event_type):
        if not subscriber.asynchronous:
            _notify_subscriber(event, subscriber)
        else:
            _async_notify_subscriber(event, subscriber)


def _get_subscribers(event_type: Type[Event]) -> list[EventSubscriber]:
    available_subscribers = ez.plugin(EventPlugin).subscribers
    return [
        subscriber
        for subscriber in available_subscribers
        if issubclass(event_type, subscriber.event_type)
    ]


def _notify_subscriber(event: Event, subscriber: EventSubscriber):
    subscriber.callback(event)


def _async_notify_subscriber(event: Event, subscriber: EventSubscriber):
    thread = Thread(target=_notify_subscriber, args=(event, subscriber), daemon=True)
    thread.start()




import enum
import inspect
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Type, TypeVar

from easy_pysi import signal
from easy_pysi.cli import run_cli
from easy_pysi.configuration import ConfigurationPlugin
from easy_pysi.event import Event, emit, EventPlugin
from easy_pysi.loop import LoopPlugin
from easy_pysi.plugin import Plugin
from easy_pysi.provider import ProviderPlugin
from easy_pysi.utils import require

logger = logging.getLogger(__name__)

T = TypeVar('T')
PluginType = TypeVar('PluginType', bound=Plugin)


class AppState(enum.Enum):
    STOPPED = 1
    INITIALIZING = 2
    STARTING = 3
    STARTED = 4
    STOPPING = 5


class App:
    modules: list[ModuleType]
    state: AppState = AppState.STOPPED
    root_directory: Path = Path(os.getcwd()).resolve()
    dotenv_path: Path = Path(os.getcwd(), '.env').resolve()
    plugins: list[Plugin] = [
        EventPlugin(),
        LoopPlugin(),
        ConfigurationPlugin(),
        ProviderPlugin(),
    ]

    def start(self):
        require(self.state == AppState.STOPPED, f"Can't start application, current state: {self.state}]")

        logger.info('EZ INIT')
        logging.basicConfig(level=logging.INFO)  # TODO: where ?
        self.state = AppState.INITIALIZING
        for plugin in self.plugins:
            plugin.init(self)

        logger.info('EZ Start')
        self.state = AppState.STARTING
        for plugin in self.plugins:
            plugin.start()

        self.state = AppState.STARTED
        applications.append(self)
        emit(AppStarted(self))

    def stop(self):
        require(self.state == AppState.STARTED, f"Can't stop application, current state: {self.state}")
        logger.info(f'Stopping {self}')
        self.state = AppState.STOPPING
        for plugin in self.plugins:
            plugin.stop()

        applications.remove(self)
        self.state = AppState.STOPPED

    def get_plugin(self, plugin_type: Type[PluginType]) -> PluginType:
        for plugin in self.plugins:
            if isinstance(plugin, plugin_type):
                return plugin
        raise RuntimeError(f'No Plugin found for: {plugin_type}')

    def is_available(self, object) -> bool:
        object_module = inspect.getmodule(object)
        return object_module in self.modules

    def run(self, auto_start=True):
        if auto_start and self.state == AppState.STOPPED:
            self.start()
        run_cli()


@dataclass
class AppStarted(Event):
    app: App


applications: list[App] = []


def current_app():
    return applications[-1]


def plugin(plugin_type: Type[PluginType]) -> PluginType:
    return current_app().get_plugin(plugin_type)


def shutdown(exit_code=0):
    stop_all()
    exit(exit_code)


def stop_all():
    logger.info('Stopping all applications')
    # [:] is used to iterate over a copy of applications (since app.stop() will remove app from applications)
    for app in applications[:]:
        app.stop()


signal.sigint_callback = shutdown




from dataclasses import dataclass
from typing import TypeVar, Type, Callable

from easy_pysi.utils import require
from easy_pysi.plugin import Plugin
import easy_pysi as ez

T = TypeVar('T')
ProviderFactory = Callable[[], T]


@dataclass
class Provider:
    type: Type[T]
    factory: ProviderFactory
    singleton: bool


_all_providers: list[Provider] = []


class ProviderPlugin(Plugin):
    providers: dict[Type[T], Provider] = {}
    singletons: dict[Type[T], T] = {}

    def start(self):
        self.providers = {
            provider.type: provider
            for provider in _all_providers
            if self.app.is_available(provider.factory)
        }


def provide(type: Type[T], singleton: bool = False):
    # TODO: auto detect async function?
    def decorator(func):
        _all_providers.append(Provider(type, func, singleton))
        return func
    return decorator


def get(type: Type[T]) -> T:  # TODO: other name
    plugin = ez.plugin(ProviderPlugin)
    require(type in plugin.providers, f"No provider found for {type}")
    provider = plugin.providers[type]

    # TODO you can write better
    if not provider.singleton:
        return provider.factory()
    elif type not in plugin.singletons:
        instance = provider.factory()
        plugin.singletons[type] = instance
        return instance
    else:
        return plugin.singletons[type]



import os
from datetime import datetime, date
from typing import Union

import pendulum
from dotenv import load_dotenv

from easy_pysi.plugin import Plugin

SupportedTypes = Union[str, int, float, bool, datetime, date]  # TODO: Bad typing
# TODO: dict and list


class ConfigurationPlugin(Plugin):
    def init(self, app):
        super().init(app)
        load_dotenv(app.dotenv_path)


def config(key: str, type: SupportedTypes = str, default=None, raise_if_not_found=False):
    raw = os.getenv(key)

    if raw is None and raise_if_not_found:
        raise ConfigNotFoundError(f'Config {key} not found')
    elif raw is None:
        return default

    if type == str:
        return raw
    elif type == int:
        return int(raw)
    elif type == float:
        return float(raw)
    elif type == bool:
        return raw.lower() == 'true' or raw == '1'
    elif type == datetime:
        return pendulum.parse(raw)
    elif type == date:
        return pendulum.parse(raw).date()
    else:
        raise TypeNotSupportedError(f'Type {type} is not supported for configuration')


class ConfigNotFoundError(Exception):
    pass


class TypeNotSupportedError(Exception):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easy_pysi']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.85.0,<0.86.0',
 'pendulum>=2.1.2,<3.0.0',
 'requests>=2.28.0,<3.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer>=0.6.1,<0.7.0',
 'uvicorn[standard]>=0.18.3,<0.19.0']

entry_points = \
{'console_scripts': ['ez = easy_pysi.console:main']}

setup_kwargs = {
    'name': 'easy-pysy',
    'version': '0.0.2',
    'description': '',
    'long_description': None,
    'author': 'Pierre Merienne',
    'author_email': 'pierre.merienne@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
import threading
import time

from typing import Optional
from uuid import uuid4

logger = logging.getLogger(__name__)


class Interval:
    def __init__(self, interval_ms, callback, on_error_callback, daemon=True, *args, **kwargs):
        self.interval_ms = interval_ms
        self.callback = callback
        self.on_error_callback = on_error_callback
        self.daemon = daemon
        self.args = args
        self.kwargs = kwargs

        self.stop_event = threading.Event()
        self.thread = threading.Thread(target=self._run, daemon=self.daemon)

    def _run(self):
        next_time = time.time() + self.interval_seconds
        while not self.stop_event.wait(next_time - time.time()):
            try:
                next_time += self.interval_seconds
                self.callback(*self.args, **self.kwargs)
            except BaseException as exc:
                self.on_error_callback(exc)

    def start(self):
        if self.thread.is_alive():
            raise RuntimeError('Interval already started')
        self.thread.start()

    def stop(self):
        if not self.thread.is_alive():
            raise RuntimeError('Thread already stopped')
        self.stop_event.set()

    @property
    def interval_seconds(self) -> float:
        return self.interval_ms / 1000.0


def require(condition: bool, message: Optional[str] = None):
    if not condition:
        raise RequirementError(message)


class RequirementError(Exception):
    pass


def uuid() -> str:
    return str(uuid4())


def tri_wave(min_value: int, max_value: int, step: int = 1):
    while True:
        yield from range(min_value, max_value, step)
        yield from range(max_value, min_value, -step)


def float_range(start: float, stop: float, step: float = 1.0, decimals: int = 2):
    for i in range(int(start / step), int(stop / step)):
        yield round(i * step, ndigits=decimals)



import logging
from dataclasses import dataclass
from typing import Callable, Optional

import easy_pysi
from easy_pysi.plugin import Plugin
from easy_pysi.utils import Interval


logger = logging.getLogger(__name__)


@dataclass
class Loop:
    interval_ms: 1000
    callback: Callable
    stop_app_on_error: bool
    auto_start: bool
    interval: Optional[Interval] = None

    def start(self):
        self.interval = Interval(self.interval_ms, self.callback, self.on_error)
        self.interval.start()

    def stop(self):
        self.interval.stop()
        self.interval = None

    def on_error(self, exception: BaseException):
        logger.error(f'Loop execution failed: {exception}')
        if self.stop_app_on_error:
            easy_pysi.current_app().stop()

    @property
    def running(self):
        return self.interval is not None


class LoopPlugin(Plugin):
    loops: list[Loop] = []

    def start(self):
        # Init loops
        self.loops = [
            loop
            for loop in _ALL_LOOPS
            if self.app.is_available(loop.callback)
        ]

        for loop in self.loops:
            if loop.auto_start:
                logger.debug(f'Starting {loop}')
                loop.start()

    def stop(self):
        for loop in self.loops:
            if loop.running:
                loop.stop()

    def get_loop(self, callback: Callable) -> Optional[Loop]:
        # TODO: nice to have :
        # ez._(self.loops).get(lambda loop: loop.callback == callback)
        for loop in self.loops:
            if loop.callback == callback:
                return loop


_ALL_LOOPS: list[Loop] = []


def loop(every_ms: float, stop_app_on_error=True, auto_start=True):
    def decorator(func):
        loop = Loop(every_ms, func, stop_app_on_error, auto_start)
        _ALL_LOOPS.append(loop)
        return func
    return decorator
